//
//  ContentView.swift
//  CoreImageMystery
//
//  Created by Marc Palmer on 09/05/2024.
//

import SwiftUI
import ImageIO
import CoreImage
import CoreImage.CIFilterBuiltins
import UIKit

// Use RGB init for these to force renderer to use sRGB instead of grayscale, saves hassle
fileprivate let fgColour = UIColor(red: 0, green: 0, blue: 0, alpha: 1)
fileprivate let bgColour = UIColor(red: 1, green: 1, blue: 1, alpha: 1)

fileprivate let sRGBColourSpace = CGColorSpace(name: CGColorSpace.sRGB)!
fileprivate let deviceRGBColourspace = CGColorSpaceCreateDeviceRGB()
fileprivate let linearSRGBColourSpace = CGColorSpace(name: CGColorSpace.linearSRGB)!

fileprivate let size = CGSize(width: 40*values.count, height: 100)
fileprivate let values = [0, 16, 32, 48, 64, 80, 96, 112, 128, 144, 160, 176, 192, 208, 224, 240, 255]
fileprivate let cellWidth: Int = Int((size.width/CGFloat(values.count)).rounded(.down))

fileprivate let contextWithColourManagement = CIContext(options: [.outputColorSpace: sRGBColourSpace])
fileprivate let contextNoColourManagement = CIContext(options: [.workingColorSpace: NSNull()])
fileprivate let contextNoColourManagementsRGBOut = CIContext(options: [.workingColorSpace: NSNull(), .outputColorSpace: sRGBColourSpace])

class SampleImages {
    /// This is the raw generated gradient in rgb with alpha 1
    static var sourceImage: UIImage = { makeSourceImage(inAlpha: false) }()

    /// This is the gradient put into the alpha channel instead
    static var sourceImageInAlpha: UIImage = { makeSourceImage(inAlpha: true) }()

    /// The problem one!
    ///
    /// This is the raw generated gradient in rgb with alpha 1
    static var sourceImageInAlphaCoreImage: UIImage = {
        let image = makeSourceImage(inAlpha: true)
        let ciImage = CIImage(image: image)!
        let backgroundImage = CIImage(color: CIColor(color: bgColour))
        let composited = ciImage.composited(over: backgroundImage)
        let rect = CGRect(origin: .zero, size: size)
        // With RGBA8 and colorSpace nil this works - without this, the colour processing is wrong
        let result = contextWithColourManagement.createCGImage(composited, from: rect)!
        print("A: Input image colourSpace: \(String(describing: image.cgImage!.colorSpace))")
        print("A: CI composite result colourSpace: \(String(describing: result.colorSpace))")
        return UIImage(cgImage: result, scale: 1, orientation: .up)
    }()

    /// The "works" but only because it skips colour conversion on render.
    static var sourceImageInAlphaCoreImageNoOutputColorSpace: UIImage = {
        let image = makeSourceImage(inAlpha: true)
        let ciImage = CIImage(image: image)!
        let backgroundImage = CIImage(color: CIColor(color: bgColour))
        let composited = ciImage.composited(over: backgroundImage)
        let rect = CGRect(origin: .zero, size: size)
        // This is very strange. If you pass *any* colorSpace here, the result is wrong — even with deviceRGB -
        // which is very confusing as the image returned if you pass nil is in `deviceRGB` so it implies
        // that internally it thinks the image is in device RGB already, but tries to apply the conversion again?
        let result = contextWithColourManagement.createCGImage(composited,
                                                               from: rect,
                                                               format: .RGBA8,
                                                               colorSpace: nil)!
        print("B: Input image colourSpace: \(String(describing: image.cgImage!.colorSpace))")
        print("B: CI composite result colourSpace: \(String(describing: result.colorSpace))")
        return UIImage(cgImage: result, scale: 1, orientation: .up)
    }()

    /// Generate a 0 -> White gradient of N steps, including 0 and 255.
    /// Uses CoreGraphics exclusively.
    static func makeSourceImage(inAlpha: Bool) -> UIImage {
        let format = UIGraphicsImageRendererFormat()
        format.scale = 1

        let renderer = UIGraphicsImageRenderer(size: size, format: format)
        let result = renderer.image { context in
            values.enumerated().forEach { index, level in
                let r = CGRect(origin: .init(x: index * cellWidth, y: 0),
                               size: .init(width: CGFloat(cellWidth), height: size.height))
                let amount = CGFloat(level)/255
                let colour: CGColor
                if inAlpha {
                    colour = fgColour.withAlphaComponent(amount).cgColor
                } else {
                    colour = CGColor(srgbRed: amount, green: amount, blue: amount, alpha: 1)
                }
                UIColor(cgColor: colour).setFill()
                context.fill(r)
            }
        }
        return result
    }
}

fileprivate func colourLabels(for image: UIImage) -> [(x: UInt, text: String)] {
    let pixels = ImageAccess(image)!
    return values.enumerated().map { index, level in
        let x: UInt = UInt(cellWidth * index)
        let colour: RGBAColor = pixels[x, 0]
        let text = """
r:\(Int(colour.r*255))
g:\(Int(colour.g*255))
b:\(Int(colour.b*255))
a:\(Int(colour.a*255))
"""
        return (x, text)
    }
}

struct CompositingContentView: View {
    var body: some View {
        ScrollView(.vertical) {
            VStack(spacing: 30) {
                Text("""
Compositing an sRGB input image in Core Image appears to misapply the alpha channel when any conversion to working color space is performed internally. Turning off colour management "fixes" it but of course creates new problems.

Notice the alpha gradient "changes its curve" after compositing over a solid colour.
""")
                .padding()
    
                sample(SampleImages.sourceImage,
                       "✅ Gradient image generated using 0→1 in rgb with a=1, sRGB", showBG: false)

                VStack {
                    sample(SampleImages.sourceImageInAlpha,
                           "✅ Gradient image rendered instead in alpha rgba(0,0,0,0→1), raw (not composited over anything), sRGB", showBG: true)
                    Text("Alpha channel values from the generated sample image 👆")
                }
                
                // HINT: Comment out the samples you don't need when using CI_PRINT_TREE so you can see just the
                // outputs that are bad e.g. sample "A"
                VStack {
                    sample(SampleImages.sourceImageInAlphaCoreImage,
                           "A: ❌ CoreImage: rgba(0,0,0,0→1) srcOver rgba(1,1,1,1). In: sRGB, Out: sRGB", showBG: false)
                    Text("Alpha blending with the white changes too slowly and then too quickly at the end 👆")
                }

                VStack {
                    sample(SampleImages.sourceImageInAlphaCoreImageNoOutputColorSpace,
                           "B: ✅ CoreImage: rgba(0,0,0,0→1) srcOver rgba(1,1,1,1). In: sRGB, Out: \"nil\" (defaults to device RGB)", showBG: false)
                    Text("Alpha blending works fine if the CI_PRINT_TREE shows it skipped all conversion after compositing 👆")
                }
            }
            .frame(maxWidth: .infinity)
        }
    }

    func sample(_ image: UIImage, _ label: String, showBG: Bool) -> some View {
        VStack {
            Text(label)
            Image(uiImage: image)
            .aspectRatio(contentMode: .fit)
            .overlay {
                ZStack(alignment: .bottom) {
                    let labels = colourLabels(for: image)
                    ForEach(labels, id: \.x) { (x, label) in
                        Text(label)
                        .position(.init(x: CGFloat(x)+20, y: 135))
                    }
                }
                .font(.caption)
                .frame(maxWidth: .infinity, maxHeight: .infinity, alignment: .bottom)
            }
            .background {
                if showBG {
                    Color.white
                    .colorEffect(ShaderLibrary.bundle(.main).checkerboard(.float(10),
                                                                                   .color(.init(white: 0.9))))
                }
            }
            .padding(.bottom, 80)
        }
    }
}

#Preview("iPad") {
    CompositingContentView()
}
