//
//  ImageAccess.swift
//
//  Created by Marc Palmer on 05/08/2023.
//

import Foundation
import UIKit

public struct RGBAColor: Equatable {
    internal init(_ r: CGFloat, _ g: CGFloat, _ b: CGFloat, _ a: CGFloat) {
        self.r = r
        self.g = g
        self.b = b
        self.a = a
    }
    
    public let r: CGFloat
    public let g: CGFloat
    public let b: CGFloat
    public let a: CGFloat
}

public class ImageAccess {
    public let image: UIImage
    private let cgImage: CGImage
    private let colorSpace: CGColorSpace?
    private let dataProvider: CGDataProvider
    private let providerData: CFData
    private let data: UnsafePointer<UInt8>
    
    public init?(_ image: UIImage) {
        self.image = image

        guard let cgImage = image.cgImage else {
            return nil
        }
        self.cgImage = cgImage
        self.colorSpace = cgImage.colorSpace

        guard let dataProvider = cgImage.dataProvider else {
            return nil
        }
        self.dataProvider = dataProvider

        guard let providerData = dataProvider.data else {
            return nil
        }
        self.providerData = providerData

        guard let data = CFDataGetBytePtr(providerData) else {
            return nil
        }
        self.data = data
    }
    
    func getYExtentOf(pixelsMatchingColour colour: CGColor, atX x: UInt, startingAtY startY: UInt) -> (min: UInt, max: UInt) {
        var result: (min: UInt, max: UInt) = (startY, startY)
        let x = UInt(x)
        for y in stride(from: startY, to: 0, by: -1) {
            if self[x, y].cgColor != colour {
                break
            }
            result.min = y
        }

        for y in startY + 1..<UInt(image.size.height) {
            if self[x, y].cgColor != colour {
                break
            }
            result.max = y + 1
        }
        
        return result
    }

    @inline(__always)
    public subscript (x: UInt, y: UInt) -> UIColor {
        precondition(x >= 0 && x < Int(image.size.width) && y >= 0 && y < Int(image.size.height))
        precondition(cgImage.bitsPerComponent == 8, "Only 8 bit images supported")

        var numberOfComponents = UInt(colorSpace?.numberOfComponents ?? 1)
        if cgImage.alphaInfo != .none && cgImage.alphaInfo != .alphaOnly {
            numberOfComponents += 1
        }
        let memoryOffset = Int((UInt(cgImage.bytesPerRow) * y) + (x * numberOfComponents))

//            print("Get pixel \(x), \(y)")
        if numberOfComponents == 4 {
            let r = CGFloat(data[memoryOffset]) / 255.0
            let g = CGFloat(data[memoryOffset + 1]) / 255.0
            let b = CGFloat(data[memoryOffset + 2]) / 255.0
            let a = CGFloat(data[memoryOffset + 3]) / 255.0

            return UIColor(red: r, green: g, blue: b, alpha: a)
        } else if numberOfComponents == 1 {
            if cgImage.alphaInfo == .alphaOnly {
                let a = CGFloat(data[memoryOffset]) / 255.0
                return UIColor(white: 0, alpha: a)
            } else {
                let g = CGFloat(data[memoryOffset]) / 255.0
                return UIColor(white: g, alpha: 1)
            }
        } else {
            fatalError("Cursed format")
        }
    }

    @inline(__always)
    public subscript (x: UInt, y: UInt) -> RGBAColor {
        precondition(x >= 0 && x < Int(image.size.width) && y >= 0 && y < Int(image.size.height))
        guard cgImage.bitsPerComponent == 8 else {
            return .init(1, 1, 1, 1)
        }

        var numberOfComponents = UInt(colorSpace?.numberOfComponents ?? 1)
        if cgImage.alphaInfo != .none && cgImage.alphaInfo != .alphaOnly {
            numberOfComponents += 1
        }
        let memoryOffset = Int((UInt(cgImage.bytesPerRow) * y) + (x * numberOfComponents))

//            print("Get pixel \(x), \(y)")
        if numberOfComponents == 4 {
            let r = CGFloat(data[memoryOffset]) / 255.0
            let g = CGFloat(data[memoryOffset + 1]) / 255.0
            let b = CGFloat(data[memoryOffset + 2]) / 255.0
            let a = CGFloat(data[memoryOffset + 3]) / 255.0

            return .init(r, g, b, a)
        } else if numberOfComponents == 1 {
            if cgImage.alphaInfo == .alphaOnly {
                let a = CGFloat(data[memoryOffset]) / 255.0
                return .init(0, 0, 0, a)
            } else {
                let g = CGFloat(data[memoryOffset]) / 255.0
                return .init(g, g, g, 1)
            }
        } else {
            fatalError("Cursed format")
        }
    }

    @inline(__always)
    public func maskValue(atX x: UInt, y: UInt) -> UInt8 {
        precondition(x >= 0 && x < Int(image.size.width) && y >= 0 && y < Int(image.size.height))
        precondition(cgImage.bitsPerComponent == 8, "Only 8 bit images supported")

        var numberOfComponents = UInt(colorSpace?.numberOfComponents ?? 1)
        if cgImage.alphaInfo != .none && cgImage.alphaInfo != .alphaOnly {
            numberOfComponents += 1
        }
        let memoryOffset = Int((UInt(cgImage.bytesPerRow) * y) + (x * numberOfComponents))

//            print("Get pixel \(x), \(y)")
        // What about ARGB?
        
        // RGBA
        if numberOfComponents == 4 {
            let a = data[memoryOffset + 3]

            return a
        } else if numberOfComponents == 1 {
            // Alpha only
            if cgImage.alphaInfo == .alphaOnly {
                let a = data[memoryOffset]
                return a
            } else {
                // Gray only
                let g = data[memoryOffset]
                return g
            }
        } else {
            fatalError("Cursed format")
        }
    }
}
